# Copyright (c) 2015 Scality
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import contextlib
import errno
import gzip
import httplib
import json
import os
import os.path
import pickle
import shutil
import tempfile
import time
import unittest

import eventlet

import swift_scality_backend.server

import swift.account.server
import swift.container.server
import swift.common.ring
import swift.common.utils
import swift.proxy.server

# Pre storage-policy compatibility
try:
    import swift.common.storage_policy
    HAS_STORAGE_POLICY = True
except ImportError:
    HAS_STORAGE_POLICY = False


class TestRangeHeaders(unittest.TestCase):
    """
    Test outgoing range headers to sproxyd from the object server on a partial
    get of an object. See
    https://github.com/scality/ScalitySproxydSwift/issues/121 for more details.
    """

    def setUp(self):
        if HAS_STORAGE_POLICY:
            # Clear out any previously set storage policies
            swift.common.storage_policy.reload_storage_policies()

        # Test app servers and associated sockets
        self.sockets = []
        self.servers = []

        # Bind proxy, account, container, and object server to a free port
        bind_address = ('127.0.0.1', 0)

        # Directory holding account, container, and object rings
        self.swift_dir = tempfile.mkdtemp()

        # Setup mocked sproxyd server
        sproxyd_socket = eventlet.listen(bind_address)
        self.sockets.append(sproxyd_socket)
        sproxyd = eventlet.spawn(eventlet.wsgi.server, sproxyd_socket,
                                 self.sproxyd_mock)
        self.servers.append(sproxyd)

        self.sproxyd_path = '/proxy/chord'
        conf = {
            'swift_dir': self.swift_dir,
            'devices': self.swift_dir,
            'mount_check': 'false',
            # Scality sproxyd config options
            'sproxyd_host': '%s:%d' % sproxyd_socket.getsockname(),
            'sproxyd_path': self.sproxyd_path,
        }

        # Hashes for the ring hashing algorithm when determining data placement
        swift.common.utils.HASH_PATH_SUFFIX = 'foo'
        swift.common.utils.HASH_PATH_PREFIX = 'bar'

        # Start swift proxy with auxiliary components.
        controllers = (
            ('account', swift.account.server.AccountController(conf)),
            ('container', swift.container.server.ContainerController(conf)),
            ('object', swift_scality_backend.server.ObjectController(conf)),
        )
        for server_name, app in controllers:
            self.setup_ring(server_name, app, bind_address)

        # Setup swift proxy
        proxy_socket = eventlet.listen(bind_address)
        self.sockets.append(proxy_socket)
        proxy = swift.proxy.server.Application(conf)
        proxy_server = eventlet.spawn(eventlet.wsgi.server, proxy_socket, proxy)
        self.servers.append(proxy_server)

        # Create test account
        account = 'a'
        partition, acc_nodes = proxy.account_ring.get_nodes(account)
        acc_node = acc_nodes.pop()
        acc_path = '/%s/%d/%s' % (acc_node['device'], partition, account)
        acc_headers = {
            'X-Timestamp': swift.common.utils.normalize_timestamp(time.time()),
        }
        acc_conn = httplib.HTTPConnection(acc_node['ip'], acc_node['port'])
        acc_conn.request('PUT', acc_path, headers=acc_headers)
        if acc_conn.getresponse().status != 201:
            raise Exception("Unable to setup test account")

        # Create test container
        self.proxy_host, self.proxy_port = proxy_socket.getsockname()
        proxy_conn = httplib.HTTPConnection(self.proxy_host, self.proxy_port)
        proxy_conn.request('PUT', '/v1/a/c/')
        if proxy_conn.getresponse().status != 201:
            raise Exception("Unable to setup test container")

    def tearDown(self):
        # Reset monkey patched hash paths
        swift.common.utils.HASH_PATH_SUFFIX = ''
        swift.common.utils.HASH_PATH_PREFIX = ''

        # Stop wsgi servers
        for wsgi_server in self.servers:
            wsgi_server.kill()

        for socket in self.sockets:
            socket.close()

        # Clean-up object rings
        shutil.rmtree(self.swift_dir, ignore_errors=1)

    def _get(self, path, headers):
        """
        Send a HTTP GET request to the swift test proxy.

        :param path: Request path
        :type path: str
        :param headers: Request headers
        :return: A tuple with response status, headers, and body.
        """
        proxy_conn = httplib.HTTPConnection(self.proxy_host, self.proxy_port)
        proxy_conn.request('GET', path, headers=headers)
        r = proxy_conn.getresponse()

        return r.status, dict(r.getheaders()), r.read()

    def test_range_from_byte(self):
        self.response = '_' * 100

        # Get last 95 bytes
        headers = {'Range': 'bytes=5-'}
        status, response_headers, _ = self._get('/v1/a/c/o', headers)
        self.assertEqual(status, 206)
        self.assertEqual(self.sproxyd_request_headers['RANGE'], 'bytes=5-99')
        self.assertEqual(response_headers['content-range'], 'bytes 5-99/100')
        self.assertEqual(response_headers['content-length'], '95')

    def test_range_last_bytes(self):
        self.response = '_' * 100

        # Get last 5 bytes
        headers = {'Range': 'bytes=-5'}
        status, response_headers, _ = self._get('/v1/a/c/o', headers)
        self.assertEqual(status, 206)
        self.assertEqual(self.sproxyd_request_headers['RANGE'], 'bytes=95-99')
        self.assertEqual(response_headers['content-range'], 'bytes 95-99/100')
        self.assertEqual(response_headers['content-length'], '5')

    def test_range_explicit_first_bytes(self):
        self.response = '_' * 100

        # Get first 5 bytes
        headers = {'Range': 'bytes=0-4'}
        status, response_headers, _ = self._get('/v1/a/c/o', headers)
        self.assertEqual(status, 206)
        self.assertEqual(self.sproxyd_request_headers['RANGE'], 'bytes=0-4')
        self.assertEqual(response_headers['content-range'], 'bytes 0-4/100')
        self.assertEqual(response_headers['content-length'], '5')

    def test_range_explicit_intermediate(self):
        self.response = '_' * 100

        headers = {'Range': 'bytes=20-40'}
        status, response_headers, _ = self._get('/v1/a/c/o', headers)
        self.assertEqual(status, 206)
        self.assertEqual(self.sproxyd_request_headers['RANGE'], 'bytes=20-40')
        self.assertEqual(response_headers['content-range'], 'bytes 20-40/100')
        self.assertEqual(response_headers['content-length'], '21')

    def test_range_explicit_end(self):
        self.response = '_' * 100

        # Get last 5 bytes
        headers = {'Range': 'bytes=95-99'}
        status, response_headers, _ = self._get('/v1/a/c/o', headers)
        self.assertEqual(status, 206)
        self.assertEqual(self.sproxyd_request_headers['RANGE'], 'bytes=95-99')
        self.assertEqual(response_headers['content-range'], 'bytes 95-99/100')
        self.assertEqual(response_headers['content-length'], '5')

    def test_range_exceed_length(self):
        self.response = '_' * 100

        # Exceed object length by one
        headers = {'Range': 'bytes=95-100'}
        status, response_headers, _ = self._get('/v1/a/c/o', headers)
        self.assertEqual(status, 206)
        self.assertEqual(self.sproxyd_request_headers['RANGE'], 'bytes=95-99')
        self.assertEqual(response_headers['content-range'], 'bytes 95-99/100')
        self.assertEqual(response_headers['content-length'], '5')

    def test_range_first_byte(self):
        self.response = '_' * 100

        # Get first byte
        headers = {'Range': 'bytes=0-0'}
        status, response_headers, _ = self._get('/v1/a/c/o', headers)
        self.assertEqual(status, 206)
        self.assertEqual(self.sproxyd_request_headers['RANGE'], 'bytes=0-0')
        self.assertEqual(response_headers['content-range'], 'bytes 0-0/100')
        self.assertEqual(response_headers['content-length'], '1')

    def setup_ring(self, server_name, app, bind_address):
        """
        Setup a single replica ring.
        """
        # Device definition
        server_socket = eventlet.listen(bind_address)
        self.sockets.append(server_socket)
        server_ip, server_port = server_socket.getsockname()
        dev = {
            'id': 0,
            'zone': 0,
            'device': 'sda1',
            'ip': server_ip,
            'port': server_port,
        }

        # Prepare ring directory
        try:
            os.makedirs(os.path.join(self.swift_dir, dev['device']))
        except OSError as e:
            # Ignore if directory already exists
            if e.errno != errno.EEXIST:
                raise

        # Compose ring metadata
        replica2part2dev = [[0, 0, 0, 0]]
        part_shift = 30
        ring = swift.common.ring.RingData(replica2part2dev, [dev], part_shift)
        path = '%s/%s.ring.gz' % (self.swift_dir, server_name)
        with contextlib.closing(gzip.GzipFile(path, 'wb')) as f:
            pickle.dump(ring, f)

        wsgi_server = eventlet.spawn(eventlet.wsgi.server, server_socket, app)
        self.servers.append(wsgi_server)

    def sproxyd_mock(self, env, start_response):
        method = env['REQUEST_METHOD']
        path = env['PATH_INFO']

        if method == 'GET' and path == '%s/.conf' % self.sproxyd_path:
            # Return configuration parameter expected by failure detector
            response = json.dumps({
                'by_path_enabled': True,
            }, indent=2)
        else:
            response = self.response

        # Expected metadata (xattr) headers by swift
        metadata = {
            'X-Timestamp': swift.common.utils.normalize_timestamp(time.time()),
            'Content-Length': len(response),
            'ETag': 'some_hash',
        }

        headers = {
            'Content-Type': 'application/json',
            'X-Scal-Usermd': base64.b64encode(pickle.dumps(metadata)),
            'Content-Length': len(response),
        }

        self.sproxyd_request_headers = dict(
            (k[len('HTTP_'):], v)
            for k, v in env.iteritems()
            if k.startswith('HTTP_')
        )

        start_response('200 OK', headers.items())
        return [response]



# Copyright (c) 2010-2013 OpenStack, LLC.
# Copyright (c) 2014 Scality
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Scality Sproxyd Object Server for Swift """

import errno

import swift.common.bufferedhttp
import swift.common.exceptions
import swift.common.http
import swift.obj.server

import scality_sproxyd_client.sproxyd_client

import swift_scality_backend.diskfile
import swift_scality_backend.http_utils
import swift_scality_backend.policy_configuration
import swift_scality_backend.utils

POLICY_STUB = object()
FRAG_PREFS_STUB = object()


class ScalityDiskFileRouter(object):
    """
    Replacement for Swift's DiskFileRouter object.
    Always returns Scality's DiskFileManager implementation.

    Copied from SwiftOnFile::obj/server.py
    """
    def __init__(self, *args, **kwargs):
        self.manager_cls = swift_scality_backend.diskfile.DiskFileManager(
            *args, **kwargs)

    def __getitem__(self, policy):
        return self.manager_cls


class ObjectController(swift.obj.server.ObjectController):
    """Implements the WSGI application for the Scality Object Server."""

    def __init__(self, *args, **kwargs):
        self._clients = {}
        self._conn_timeout = None
        self._read_timeout = None
        self._diskfile_mgr = None
        self._policy_configuration = None
        self._policy_0_urls = None
        self._location_preferences = None
        self._url_password = None
        self._url_username = None
        self._url_cert_bundle = None
        self._url_client_cert = None
        self._url_client_key = None

        super(ObjectController, self).__init__(*args, **kwargs)

    def setup(self, conf):
        """Class setup

        :param conf: WSGI configuration parameter
        """
        # Replaces Swift's DiskFileRouter object reference with ours.
        self._diskfile_router = ScalityDiskFileRouter(conf, self.logger)

        self._url_username = conf.get('sproxyd_url_username')
        self._url_password = conf.get('sproxyd_url_password')
        self._url_cert_bundle = conf.get('sproxyd_url_cert_bundle')
        self._url_client_cert = conf.get('sproxyd_url_client_cert')
        self._url_client_key = conf.get('sproxyd_url_client_key')

        # New style configuration expects
        # sproxyd_endpoints = http://sproxyd1:port/path1, http://sproxyd2:port/path2
        # in object-server.conf
        if conf.get('sproxyd_endpoints'):
            self._policy_0_urls = swift_scality_backend.utils.split_list(
                conf['sproxyd_endpoints'])
        # But we still support specifying sproxyd_host and sproxyd_path
        else:
            sproxyd_path = conf.get('sproxyd_path', '/proxy/chord').strip('/')
            self._policy_0_urls = [
                'http://%s/%s/' % (h, sproxyd_path)
                for h in swift_scality_backend.utils.split_list(conf['sproxyd_host'])]

        def float_or_none(value):
            return float(value) if value is not None else None

        self._conn_timeout = float_or_none(conf.get('sproxyd_conn_timeout'))
        self._read_timeout = float_or_none(conf.get('sproxyd_proxy_timeout'))

        self._diskfile_mgr = swift_scality_backend.diskfile.DiskFileManager(conf, self.logger)

        sp_path = \
            swift_scality_backend.policy_configuration.DEFAULT_CONFIGURATION_PATH
        self.logger.info('Reading storage policy configuration from %r', sp_path)
        try:
            with open(sp_path, 'r') as fd:
                self.logger.info('Parsing storage policy configuration')
                self._policy_configuration = \
                    swift_scality_backend.policy_configuration.Configuration.from_stream(fd)
        except IOError as exc:
            if exc.errno == errno.ENOENT:
                self.logger.info(
                    'No storage policy configuration found at %r', sp_path)
                self._policy_configuration = None
            else:
                self.logger.exception(
                    'Failure while reading storage policy configuration '
                    'from %r', sp_path)
                raise

        location_preferences = conf.get('scality_location_preferences')
        if location_preferences is None:
            self._location_preferences = None
        else:
            self._location_preferences = \
                swift_scality_backend.utils.split_list(location_preferences)

        self.logger.info('=== Begin swift_scality_backend configuration ===')
        self.logger.info(repr({
            'policy_0_urls': self._policy_0_urls,
            'conn_timeout': self._conn_timeout,
            'read_timeout': self._read_timeout,
            'policy_configuration': self._policy_configuration,
            'location_preferences': self._location_preferences,
        }))
        self.logger.info('=== End swift_scality_backend configuration ===')

    def _get_client_for_policy(self, policy_idx):
        '''Retrieve or create an Sproxyd client for a given storage policy

        :param policy_idx: Policy identifier
        :type policy_idx: `int`
        :return: Sproxyd client which can be used for requests in the given
                 policy
        :rtype: `scality_sproxyd_client.sproxyd_client.SproxydClient`

        :raise RuntimeError: No policies configured
        '''

        # Static arguments to pass to the `SproxydClient` constructor.
        # Note we can only add `conn_timeout` and `read_timeout` when configured
        # (i.e. not pass `None` to the constructor) because of the default
        # values in the constructor arguments.
        sproxyd_client_kwargs = {
            'logger': self.logger,
        }
        if self._url_cert_bundle is not None:
            sproxyd_client_kwargs['url_cert_bundle'] = self._url_cert_bundle
        if self._url_client_cert is not None:
            sproxyd_client_kwargs['url_client_cert'] = self._url_client_cert
        if self._url_client_key is not None:
            sproxyd_client_kwargs['url_client_key'] = self._url_client_key
        if self._url_username is not None:
            sproxyd_client_kwargs['url_username'] = self._url_username
        if self._url_password is not None:
            sproxyd_client_kwargs['url_password'] = self._url_password
        if self._conn_timeout is not None:
            sproxyd_client_kwargs['conn_timeout'] = self._conn_timeout
        if self._read_timeout is not None:
            sproxyd_client_kwargs['read_timeout'] = self._read_timeout

        # Set up a new client collection when we didn't create one before,
        # otherwise just return the old one
        if policy_idx not in self._clients:
            collection = None

            # 'Default' (old-style) configuration (no custom storage policy)
            if policy_idx == 0:
                endpoints = self._policy_0_urls

                client = scality_sproxyd_client.sproxyd_client.SproxydClient(
                    endpoints, **sproxyd_client_kwargs)

                collection = swift_scality_backend.http_utils.ClientCollection(
                    read_clients=[client], write_clients=[client])
            else:
                if not self._policy_configuration:
                    raise RuntimeError(
                        'No storage policy configuration found, but request '
                        'for policy %r' % policy_idx)

                policy = self._policy_configuration.get_policy(policy_idx)

                # Clients in a write set of a policy are assumed to be readable
                # as well (although potentially with a lower precedence). To
                # ensure we only create a single `SproxydClient` for a given set
                # of endpoints (to reduce the number of connections to those
                # endpoints, and not to duplicate the failure detector requests
                # sent to it), this map caches `SproxydClient`s based on their
                # set of endpoints for re-use.
                clients = {}

                # Iterate over all *read* endpoints and construct
                # `SproxydClient` instances accordingly, or re-use an existing
                # one.
                read_clients = []
                for endpoints in policy.lookup(
                        policy.READ, location_hints=self._location_preferences):
                    if endpoints in clients:
                        # This should only happen for very funky configurations
                        read_clients.append(clients[endpoints])
                    else:
                        client = scality_sproxyd_client.sproxyd_client.SproxydClient(
                            (endpoint.url for endpoint in endpoints),
                            **sproxyd_client_kwargs)
                        clients[endpoints] = client

                        read_clients.append(client)

                # Iterate over all *write* endpoints and construct
                # `SproxydClient` instances accordingly, or re-use an existing
                # one.
                write_clients = []
                for endpoints in policy.lookup(
                        policy.WRITE, location_hints=self._location_preferences):
                    if endpoints in clients:
                        write_clients.append(clients[endpoints])
                    else:
                        client = scality_sproxyd_client.sproxyd_client.SproxydClient(
                            (endpoint.url for endpoint in endpoints), self._url_cert_bundle,
                            self._url_client_cert, self._url_client_key, self._url_username,
                            self._url_password, self._conn_timeout, self._read_timeout,
                            self.logger)
                        clients[endpoints] = client

                        write_clients.append(client)

                collection = swift_scality_backend.http_utils.ClientCollection(
                    read_clients=read_clients, write_clients=write_clients)

            # Cache the collection for this policy
            self._clients[policy_idx] = collection

            self.logger.info(
                '=== Begin swift_scality_backend configuration for '
                'storage policy %r ===' % policy_idx)
            self.logger.info(repr(collection))
            self.logger.info(
                '=== End swift_scality_backend configuration for '
                'storage policy %r ===' % policy_idx)

        return self._clients[policy_idx]

    def get_diskfile(self, device, partition, account, container, obj,
                     policy=POLICY_STUB, frag_prefs=FRAG_PREFS_STUB, **kwargs):
        """
        Utility method for instantiating a DiskFile object supporting a
        given REST API.
        """

        if 'policy_idx' in kwargs:
            if policy is POLICY_STUB:
                policy = kwargs['policy_idx']
            else:
                raise ValueError('Both `policy_idx` and `policy` provided')

        # When `policy_idx` is not set (e.g. running Swift 1.13), the fallback
        # policy 0 should be used.
        if policy is POLICY_STUB:
            policy = 0

        # In Swift 2.3 policies are no longer defined by an integer, but by a
        # value of type `swift.common.storage_policy.BaseStoragePolicy`, which
        # has an `idx` attribute.
        # We rely on the integer only, for now.
        policy = getattr(policy, 'idx', policy)

        client_collection = self._get_client_for_policy(policy)

        return self._diskfile_mgr.get_diskfile(
            client_collection, account, container, obj)

    def REPLICATE(ctrl=None, *_args, **_kwargs):
        """Handle REPLICATE requests for the Swift Object Server.

        This is used by the object replicator to get hashes for directories.
        """
        pass


def app_factory(global_conf, **local_conf):
    """paste.deploy app factory for creating WSGI object server apps."""
    conf = global_conf.copy()
    conf.update(local_conf)
    return ObjectController(conf)
